! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                   Test for MPI_Send() to MPI_PROC_NULL
! 
! This tests the basic blocking MPI_Send operation to MPI_PROC_NULL.  No
! messages should be sent.
! 
! This test may be run in any communicator, with any data type, and with
! any non-negative message length.
! 
! The MPITEST environment provides looping over communicator size and
! type, message length, and data type.  The properties of the loops are
! encoded in configuration arrays in the file config.h .  See the
! MPITEST README for further details.
! 
! **********************************************************************
#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    COMM_SIZE
!                               number of ranks in the current communicator
      INTEGER    TYPE_COUNT
!                               number of data types to test
      INTEGER    TEST_TYPE
!                               index of current data type
      INTEGER    LENGTH_COUNT
!                               number of data lengths to test
      INTEGER    MAX_LENGTH
!                               max message length to test
      INTEGER    MAX_BYTE_LENGTH
!                               max byte length to test
      INTEGER    MAX_BUFF_LENGTH
!                               max length that will fit in buffers
      INTEGER    BYTE_LENGTH
!                               current byte length to test
      INTEGER    LENGTH
!                               current length to test 

      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    COMM
!                               MPI communicator under test
      INTEGER    I, J, K

      MPITEST_BUF_TYPE BUFFER(MAX_BUFF_SIZE+1)
      INTEGER    ERRSIZE

      MPITEST_BUF_TYPE VALUE

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
      CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

 99   FORMAT(A,INT_FMT)
 179  FORMAT(A,I6,A,I6,A,I6,A,I3,A,INT_FMT)

!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Send_null'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF


!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0

!
!     Get number of communicators, data types and message lengths to test
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
      CALL MPITEST_NUM_DATATYPES(TYPE_COUNT)
      CALL MPITEST_NUM_MESSAGE_LENGTHS(LENGTH_COUNT)

!
!     Get max message length for test
!
      CALL MPITEST_GET_MAX_MESSAGE_LENGTH(MAX_BYTE_LENGTH)

!
!     Loop through communicators
!
      DO 100 I=1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(I, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(I, COMM_TYPE)

        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              COMM_SIZE)

!       Skip to end of loop if this node is not part of current
!       communicator

        IF (COMM .EQ. MPI_COMM_NULL) GO TO 201

!
!       Loop through data types.
!
        DO 200 J=1,TYPE_COUNT

          CALL MPITEST_GET_DATATYPE(J, TEST_TYPE)
!
!         Convert the number of bytes in the maximum length message
!         into the number of elements of the current type 
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, MAX_BYTE_LENGTH, &
                MAX_LENGTH)
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, &
                 MAX_BUFF_SIZE*MPITEST_BUF_EXTENT, MAX_BUFF_LENGTH)

          IF (MAX_LENGTH .GE. (MAX_BUFF_LENGTH - 1)) &
               MAX_LENGTH = MAX_BUFF_LENGTH - 1


!
!         Loop through data lengths
!
          DO 300 K=1,LENGTH_COUNT
            CALL MPITEST_GET_MESSAGE_LENGTH(K, BYTE_LENGTH)
            CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, BYTE_LENGTH, &
                LENGTH)
            IF (LENGTH .GT. MAX_LENGTH) LENGTH = MAX_LENGTH

!
!           All nodes send a message to MPI_PROC_NULL
!
!           Print an informational message
            IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
              INFOBUF=' '
              WRITE (INFOBUF,179) 'commindex',COMM_INDEX,', commsize', &
                  COMM_SIZE, ', commtype', COMM_TYPE,  &
                  ', datatype', TEST_TYPE, ', length', LENGTH
              CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
            END IF
            LOOP_CNT = LOOP_CNT + 1

            CALL MPI_SEND(BUFFER, LENGTH, &
                   MPITEST_MPI_DATATYPES(TEST_TYPE), MPI_PROC_NULL, &
                   MPITEST_ME, COMM, IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              INFOBUF=' '
              WRITE(INFOBUF,99) 'MPI_SEND() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
              FAIL = FAIL + 1
            END IF

 300      CONTINUE

 200    CONTINUE
 201    CONTINUE

        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)

 100  CONTINUE

!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END
 
 
