#include "f90options.h"

      INTEGER FUNCTION ADDOP(INVEC, INOUTVEC, LEN, DTYPE)
! *************************************************************************
!   Commutative user Op (addition)
! *************************************************************************
      INCLUDE 'mpitestf90.h'


      INTEGER  INVEC(*)
      INTEGER  INOUTVEC(*)
      INTEGER  LEN
      INTEGER  DTYPE
      INTEGER  I

      EXTERNAL ADDOPR
      EXTERNAL ADDOPD
      EXTERNAL ADDOPC

      IF (LEN .EQ. 0) GO TO 101
      IF (DTYPE .EQ. MPI_INTEGER) THEN
        DO 100 I=1,LEN
          INOUTVEC(I) = INVEC(I) + INOUTVEC(I)
 100    CONTINUE
      ELSE IF (DTYPE .EQ. MPI_REAL) THEN
        CALL ADDOPR(INVEC, INOUTVEC, LEN)
      ELSE IF (DTYPE .EQ. MPI_DOUBLE_PRECISION) THEN
        CALL ADDOPD(INVEC, INOUTVEC, LEN)
      ELSE IF (DTYPE .EQ. MPI_COMPLEX) THEN
        CALL ADDOPC(INVEC, INOUTVEC, LEN)
      END IF

 101  CONTINUE
      ADDOP = 0

      END


      SUBROUTINE ADDOPR(INVEC, INOUTVEC, LEN)
! *************************************************************************
!   Commutative user Op (addition) for REAL
! *************************************************************************

      REAL     INVEC(*)
      REAL     INOUTVEC(*)
      INTEGER  LEN
      INTEGER  I

      DO 100 I=1,LEN
        INOUTVEC(I) = INVEC(I) + INOUTVEC(I)
 100  CONTINUE

      END


      SUBROUTINE ADDOPD(INVEC, INOUTVEC, LEN)
! *************************************************************************
!   Commutative user Op (addition) for DOUBLE PRECISION
! *************************************************************************

      DOUBLE PRECISION  INVEC(*)
      DOUBLE PRECISION  INOUTVEC(*)
      INTEGER  LEN
      INTEGER  I

      DO 100 I=1,LEN
        INOUTVEC(I) = INVEC(I) + INOUTVEC(I)
 100  CONTINUE

      END

      SUBROUTINE ADDOPC(INVEC, INOUTVEC, LEN)
! *************************************************************************
!   Commutative user Op (addition) for COMPLEX
! *************************************************************************

      COMPLEX  INVEC(*)
      COMPLEX  INOUTVEC(*)
      INTEGER  LEN
      INTEGER  I

      DO 100 I=1,LEN
        INOUTVEC(I) = INVEC(I) + INOUTVEC(I)
 100  CONTINUE

      END


      INTEGER FUNCTION INCOP(INVEC, INOUTVEC, LEN, DTYPE)
! *************************************************************************
! Non-commutative user Op.
! *************************************************************************
      INCLUDE 'mpitestf90.h'


      INTEGER  INVEC(*)
      INTEGER  INOUTVEC(*)
      INTEGER  LEN
      INTEGER  DTYPE
      INTEGER  I

      EXTERNAL INCOPR
      EXTERNAL INCOPD
      EXTERNAL INCOPC

      IF (LEN .EQ. 0) GO TO 101
      IF (DTYPE .EQ. MPI_INTEGER) THEN
        DO 100 I=1,LEN
          IF (INOUTVEC(I) .GT. INVEC(I)) THEN
            INOUTVEC(I) = INVEC(I)
          ELSE
            INOUTVEC(I) = MPITEST_NUMP + 2
          END IF
 100    CONTINUE
      ELSE IF (DTYPE .EQ. MPI_REAL) THEN
        CALL INCOPR(INVEC, INOUTVEC, LEN)
      ELSE IF (DTYPE .EQ. MPI_DOUBLE_PRECISION) THEN
        CALL INCOPD(INVEC, INOUTVEC, LEN)
      ELSE IF (DTYPE .EQ. MPI_COMPLEX) THEN
        CALL INCOPC(INVEC, INOUTVEC, LEN)
      END IF

 101  CONTINUE
      INCOP = 0
      END


      SUBROUTINE INCOPR(INVEC, INOUTVEC, LEN)
! *************************************************************************
! Non-commutative user Op for REAL.
! *************************************************************************
      INCLUDE 'mpitestf90.h'

      REAL     INVEC(*)
      REAL     INOUTVEC(*)
      INTEGER  LEN
      INTEGER  DTYPE
      INTEGER  I

      DO 100 I=1,LEN
        IF (INOUTVEC(I) .GT. INVEC(I)) THEN
          INOUTVEC(I) = INVEC(I)
        ELSE
          INOUTVEC(I) = MPITEST_NUMP + 2
        END IF
 100  CONTINUE

      END

      SUBROUTINE INCOPD(INVEC, INOUTVEC, LEN)
! *************************************************************************
! Non-commutative user Op for DOUBLE PRECISION.
! *************************************************************************
      INCLUDE 'mpitestf90.h'

      DOUBLE PRECISION  INVEC(*)
      DOUBLE PRECISION  INOUTVEC(*)
      INTEGER  LEN
      INTEGER  DTYPE
      INTEGER  I

      DO 100 I=1,LEN
        IF (INOUTVEC(I) .GT. INVEC(I)) THEN
          INOUTVEC(I) = INVEC(I)
        ELSE
          INOUTVEC(I) = MPITEST_NUMP + 2
        END IF
 100  CONTINUE

      END

      SUBROUTINE INCOPC(INVEC, INOUTVEC, LEN)
! *************************************************************************
! Non-commutative user Op for COMPLEX.
! *************************************************************************
      INCLUDE 'mpitestf90.h'

      COMPLEX  INVEC(*)
      COMPLEX  INOUTVEC(*)
      INTEGER  LEN
      INTEGER  DTYPE
      INTEGER  I

      DO 100 I=1,LEN
        IF (REAL(INOUTVEC(I)) .GT. REAL(INVEC(I))) THEN
          INOUTVEC(I) = INVEC(I)
        ELSE
          INOUTVEC(I) = CMPLX(MPITEST_NUMP + 2)
        END IF
 100  CONTINUE

      END

      SUBROUTINE SET_DEFAULT_OPS(OP_ARRAY, NUMBER)
! ************************************************************************
!  Sets up the default operation array.  Returns the number of default 
!  operations.
! ************************************************************************
      INCLUDE 'mpitestf90.h'

      INCLUDE 'externalf90.h'

      INTEGER OP_ARRAY(10)
      INTEGER NUMBER

      EXTERNAL ADDOP
      EXTERNAL INCOP

      CHARACTER*(IOSIZE) INFOBUF
      INTEGER IERR
      INTEGER ERRSIZE
      INTEGER ERR

 99   FORMAT(A,INT_FMT)


      CALL MPI_OP_CREATE(ADDOP, .TRUE., OP_ARRAY(1), IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_OP_CREATE() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF=' '
        CALL MPI_ERROR_STRING(IERR, INFOBUF, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
 
      CALL MPI_OP_CREATE(INCOP, .FALSE., OP_ARRAY(2), IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_OP_CREATE() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF=' '
        CALL MPI_ERROR_STRING(IERR, INFOBUF, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      NUMBER = 2

      END


      SUBROUTINE HAS_OP(OP, TEST_TYPE, HASOP)
! ************************************************************************
! Determines whether a particular operation may be applied to a particular 
! data type.
! ************************************************************************
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

      INTEGER OP
      INTEGER TEST_TYPE
      LOGICAL HASOP

      IF ((TEST_TYPE .EQ. MPITEST_INTEGER) .OR. &
          (TEST_TYPE .EQ. MPITEST_REAL) .OR. &
          (TEST_TYPE .EQ. MPITEST_DOUBLE_PRECISION) .OR. &
          (TEST_TYPE .EQ. MPITEST_COMPLEX)) THEN
        HASOP = .TRUE.
      ELSE
        HASOP = .FALSE.
      END IF

      END


      SUBROUTINE GET_INT_REDUCE(OP_INDEX, NUMP, ANSWER)
! ************************************************************************
! Get result of integer Reduction
! ************************************************************************
!
      INCLUDE 'mpitestf90.h'

      INCLUDE 'externalf90.h'

      INTEGER OP_INDEX, NUMP

      INTEGER I
      INTEGER INVEC(MAX_RANKS), ANSWER, INOUTVEC(MAX_RANKS)

      IF (NUMP .EQ. 1) THEN
         ANSWER = 1

      ELSE IF (NUMP .GT. 1) THEN
         INVEC(1) = 1
         INOUTVEC(1) = 2

         IF (OP_INDEX .EQ. 1) THEN
            CALL ADDOP(INVEC, INOUTVEC, 1, MPI_INTEGER)
         ELSE IF (OP_INDEX .EQ. 2) THEN
            CALL INCOP(INVEC, INOUTVEC, 1, MPI_INTEGER)
         END IF

         IF (NUMP .GT. 2) THEN
            DO 100 I = 3, NUMP
!
!              Since the operator may not be commutative
!
               INVEC(1) = INOUTVEC(1)

               INOUTVEC(1) = I

               IF (OP_INDEX .EQ. 1) THEN
                  CALL ADDOP(INVEC, INOUTVEC, 1, MPI_INTEGER)
               ELSE IF (OP_INDEX .EQ. 2) THEN
                  CALL INCOP(INVEC, INOUTVEC, 1, MPI_INTEGER)
               END IF

 100        CONTINUE

         END IF

         ANSWER = INOUTVEC(1)

      END IF

      END


      SUBROUTINE GET_REAL_REDUCE(OP_INDEX, NUMP, ANSWER)
! ************************************************************************
! Get result of integer Reduction
! ************************************************************************
!
      INCLUDE 'mpitestf90.h'

      INCLUDE 'externalf90.h'

      INTEGER OP_INDEX, NUMP

      INTEGER I
      REAL INVEC(MAX_RANKS), ANSWER, INOUTVEC(MAX_RANKS)

      IF (NUMP .EQ. 1) THEN
         ANSWER = 1

      ELSE IF (NUMP .GT. 1) THEN
         INVEC(1) = 1
         INOUTVEC(1) = 2

         IF (OP_INDEX .EQ. 1) THEN
            CALL ADDOP(INVEC, INOUTVEC, 1, MPI_REAL)
         ELSE IF (OP_INDEX .EQ. 2) THEN
            CALL INCOP(INVEC, INOUTVEC, 1, MPI_REAL)
         END IF

         IF (NUMP .GT. 2) THEN
            DO 100 I = 3, NUMP
!
!              Since the operator may not be commutative
!
               INVEC(1) = INOUTVEC(1)

               INOUTVEC(1) = I

               IF (OP_INDEX .EQ. 1) THEN
                  CALL ADDOP(INVEC, INOUTVEC, 1, MPI_REAL)
               ELSE IF (OP_INDEX .EQ. 2) THEN
                  CALL INCOP(INVEC, INOUTVEC, 1, MPI_REAL)
               END IF

 100        CONTINUE

         END IF

         ANSWER = INOUTVEC(1)

      END IF

      END


      SUBROUTINE GET_COMPLEX_REDUCE(OP_INDEX, NUMP, &
                                    ANSWER)
! ************************************************************************
! Get result of integer Reduction
! ************************************************************************
!
      INCLUDE 'mpitestf90.h'

      INCLUDE 'externalf90.h'

      INTEGER OP_INDEX, NUMP

      INTEGER I
      COMPLEX INVEC(MAX_RANKS), ANSWER, INOUTVEC(MAX_RANKS)

      IF (NUMP .EQ. 1) THEN
         ANSWER = 1

      ELSE IF (NUMP .GT. 1) THEN
         INVEC(1) = 1
         INOUTVEC(1) = 2

         IF (OP_INDEX .EQ. 1) THEN
            CALL ADDOP(INVEC, INOUTVEC, 1, MPI_COMPLEX)
         ELSE IF (OP_INDEX .EQ. 2) THEN
            CALL INCOP(INVEC, INOUTVEC, 1, MPI_COMPLEX)
         END IF

         IF (NUMP .GT. 2) THEN
            DO 100 I = 3, NUMP
!
!              Since the operator may not be commutative
!
               INVEC(1) = INOUTVEC(1)

               INOUTVEC(1) = I

               IF (OP_INDEX .EQ. 1) THEN
                  CALL ADDOP(INVEC, INOUTVEC, 1, MPI_COMPLEX)
               ELSE IF (OP_INDEX .EQ. 2) THEN
                  CALL INCOP(INVEC, INOUTVEC, 1, MPI_COMPLEX)
               END IF

 100        CONTINUE

         END IF

         ANSWER = INOUTVEC(1)

      END IF

      END


      SUBROUTINE GET_DOUBLE_REDUCE(OP_INDEX, NUMP, &
                                    ANSWER)
! ************************************************************************
! Get result of integer Reduction
! ************************************************************************
!
      INCLUDE 'mpitestf90.h'

      INCLUDE 'externalf90.h'

      INTEGER OP_INDEX, NUMP

      INTEGER I
      DOUBLE PRECISION INVEC(MAX_RANKS), ANSWER, INOUTVEC(MAX_RANKS)

      IF (NUMP .EQ. 1) THEN
         ANSWER = 1

      ELSE IF (NUMP .GT. 1) THEN
         INVEC(1) = 1
         INOUTVEC(1) = 2

         IF (OP_INDEX .EQ. 1) THEN
            CALL ADDOP(INVEC, INOUTVEC, 1, MPI_DOUBLE_PRECISION)
         ELSE IF (OP_INDEX .EQ. 2) THEN
            CALL INCOP(INVEC, INOUTVEC, 1, MPI_DOUBLE_PRECISION)
         END IF

         IF (NUMP .GT. 2) THEN
            DO 100 I = 3, NUMP
!
!              Since the operator may not be commutative
!
               INVEC(1) = INOUTVEC(1)

               INOUTVEC(1) = I

               IF (OP_INDEX .EQ. 1) THEN
                  CALL ADDOP(INVEC, INOUTVEC, 1, MPI_DOUBLE_PRECISION)
               ELSE IF (OP_INDEX .EQ. 2) THEN
                  CALL INCOP(INVEC, INOUTVEC, 1, MPI_DOUBLE_PRECISION)
               END IF

 100        CONTINUE

         END IF

         ANSWER = INOUTVEC(1)

      END IF

      END


      SUBROUTINE GET_REDUCE_ANSWER(OP_INDEX, NUMP, ANSWER, TYPE)
! ************************************************************************
! Apply the binary operation specified by op_index to the numbers 
! (0, 1, 2, ..., nump-1), and fill in the dataTamplate object based on the 
! results.  The mapping of op_index to operation is determined by the array 
! MPITEST_default_ops[], which is set at runtime in the main test code. 
! In order for the answer produced by this routine to match the
! answer generated by the MPI_Reduce() operation in the test code, the
! send buffer of process with rank "i" must have been initialized with "i". 
! 
! This routine applies the operation to both integers and to doubles, 
! in case the double and float buffers are initialized differently than the 
! integer buffers.
! ************************************************************************
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

      INTEGER OP_INDEX
      INTEGER NUMP, TYPE
      MPITEST_BUF_TYPE ANSWER

      IF (TYPE .EQ. MPITEST_INTEGER) THEN
         CALL GET_INT_REDUCE(OP_INDEX, NUMP, ANSWER)
      ELSE IF (TYPE .EQ. MPITEST_REAL) THEN
         CALL GET_REAL_REDUCE(OP_INDEX, NUMP, ANSWER)
      ELSE IF (TYPE .EQ. MPITEST_DOUBLE_PRECISION) THEN
         CALL GET_DOUBLE_REDUCE(OP_INDEX, NUMP, ANSWER)
      ELSE IF (TYPE .EQ. MPITEST_COMPLEX) THEN
         CALL GET_COMPLEX_REDUCE(OP_INDEX, NUMP, ANSWER)
      END IF

      END

