! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
!  
!                          MPI_Group_incl_err2 
!
!  This test verifies that the correct error is returned if MPI_Group_incl()
!  is called with invalid arguments.
!  
!  MPI_Group_incl error tests
!  -----------------------------------
!  1)  Call with MPI_GROUP_NULL.......................[MPI_ERR_GROUP]
!  2)  Call with negative n...........................[MPI_ERR_OTHER/ARG]
!  3)  Call with n greater then the size of group.....[MPI_ERR_OTHER/ARG]
!  4)  Call with negative rank........................[MPI_ERR_RANK]
!  5)  Call with out of range (positive) rank.........[MPI_ERR_RANK]
!  6)  Call with duplicate ranks......................[MPI_ERR_OTHER/ARG/RANK]
!  
!  In all cases, expect to receive appropriate error.
!  
!  Rank 0 will call MPI_Group_incl with negative n.
!  The resulting error code will then be checked and the corresponding
!  error class will be verified to make sure it is MPI_ERR_GROUP.
!  
!  All other rank(s) will simply do nothing.
!  
!  MPI Calls dependencies for this test:
!    MPI_Group_incl(), MPI_Init(), MPI_Finalize()
!    MPI_Error_string()
!***********************************************************************

#include "f90options.h"

      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
      CHARACTER  TEST_DESC*(*)
      PARAMETER( TEST_DESC = &
             'MPI_Group_incl: call with negative N')
!
      INTEGER PASS
      INTEGER FAIL              
!                           counts total number of PASSes/FAILures
      INTEGER IERR
      INTEGER ERR
!                           return value from MPI calls
      INTEGER ERRORCLASS
      INTEGER SIZE 
!                           error class of IERR
      CHARACTER*(IOSIZE)  INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
      CHARACTER*32   TESTNAME
!                           name of this test
      INTEGER I
!                           utility index variable
      INTEGER N
!                           count of ranks to include
      INTEGER RANKS(MAX_RANKS)
!                           ranks to include
      INTEGER GROUP, NEWGROUP
!                           group handles

!-----------------------------------------------------------------------------
 98   FORMAT(A,A)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT)
 102  FORMAT(A, INT_FMT, A, INT_FMT)
 103  FORMAT(A, INT_FMT)
 105  FORMAT(A, INT_FMT, A, INT_FMT, A)
!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,103) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
      TESTNAME='MPI_Group_incl_err2'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF
!
      PASS = 0
      FAIL = 0
! 
!  Set an errorhandler so we get control back.  
! 
      CALL MPI_ERRHANDLER_SET(MPI_COMM_WORLD, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103) 'MPI_ERRORHANDLER_SET returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!-----------------------  MPI_Comm_group  -----------------------------
!
      CALL MPI_COMM_GROUP(MPI_COMM_WORLD, GROUP, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103) 'MPI_COMM_GROUP returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
      IF(MPITEST_ME .EQ. 0) THEN
        INFOBUF= &
          ' Calling MPI_Group_incl with negative N'
        CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
!
!     Set variables for call
!
        N = -1
!
        DO 320 I = 1, MPITEST_NUMP
          RANKS(I) = 1
 320    CONTINUE
!
!-----------------------  MPI_Group_incl  -----------------------------
!
        CALL MPI_GROUP_INCL(GROUP, N, RANKS, NEWGROUP, IERR)
        IF (IERR .EQ. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF, 98) TEST_DESC, ' did not FAIL'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ELSE
          CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
          IF (ERRORCLASS .NE. MPI_ERR_ARG  .AND.   &
              ERRORCLASS .NE. MPI_ERR_OTHER) THEN
            FAIL = FAIL + 1
            WRITE(INFOBUF,101) TEST_DESC, ' Returned',  ERRORCLASS 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            WRITE(INFOBUF,105)  &
              'Expected MPI_ERR_ARG(',    MPI_ERR_ARG,   &
              ')  OR  MPI_ERR_OTHER(',  MPI_ERR_OTHER, ')' 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ELSE
            PASS = PASS + 1
            WRITE(INFOBUF,102) 'IERR = ', IERR, '  ERRORCLASS = ', &
              ERRORCLASS
            CALL MPITEST_MESSAGE(MPITEST_INFO2, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
          END IF
        END IF
      END IF

      CALL MPI_GROUP_FREE(GROUP, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103) 'MPI_GROUP_FREE returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      CALL MPITEST_REPORT(PASS, FAIL, 0, TESTNAME)
      CALL MPI_FINALIZE(IERR)
      END
