! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
! 			  Test for MPI_Comm_dup
! 
! This test verifies that MPI_Comm_dup creates a seperate commuicator,
! and that message traffic is sorted by communicator.  Each configured
! communicator is duplicated and tested.
! 
! As a consequence of this test, MPI_Comm_free, MPI_Comm_compare, 
! MPI_Comm_size, MPI_Comm_test_inter and MPI_Comm_rank are also 
! tested here.  Testing of cached information is done in the tests 
! for cached information calls.
! 
! Test history:
!    1  08/28/96     gt/jh       Original version
! 
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'
!
      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
!  Program constants (Parameters)
!
      CHARACTER*32          TESTNAME
      PARAMETER( TESTNAME = 'MPI_COMM_DUP')
!
!  General variables that occur in all programs
!
      INTEGER    CNUMBR
!                               index for communications loop
      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    TEST_NUMP
!                               number of ranks in the current communicator
      INTEGER    LOOP_COUNT
!                               counts total number of failures, loops
      INTEGER    COMM
!                               MPITEST  communicator under test
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ERRORCLASS
      INTEGER    SIZE 
!                               error class of IERR
      INTEGER    ERRSIZE
!                               length of error message
      INTEGER    FAIL              
      INTEGER    VERIFY
!                           counts total number of PASSes/FAILures/Verifies
      CHARACTER*(IOSIZE)    INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
!
!  Variables specific to this program
!
      LOGICAL    INTERCOMM
      LOGICAL    INTERCOMM2
!                               intercommunicator flag
      INTEGER    RESULT
!                               communicator test output
      INTEGER    RESULTDUP
!                               communicator test output
      INTEGER    BUFFER1
      INTEGER    BUFFER2
      INTEGER    BUFFER3
!                               Message buffers for ISENDs
      INTEGER    COMM2
!                               for dup'ing communicator
      INTEGER    REQUEST1
      INTEGER    REQUEST2
      INTEGER    REQUEST3
!                               Communication request handles for ISENDs
      INTEGER    STATUS1(MPI_STATUS_SIZE)
      INTEGER    STATUS2(MPI_STATUS_SIZE)
      INTEGER    STATUS3(MPI_STATUS_SIZE)
!                               Status arrays for WAITs
!-----------------------------------------------------------------------------
 98   FORMAT(A,A )
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT, A, INT_FMT)
 102  FORMAT(A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT)
 103  FORMAT(A, INT_FMT, A, A, A, INT_FMT)
 104  FORMAT(A, A, A, INT_FMT, A)
 105  FORMAT(A, INT_FMT, A, A, INT_FMT, A)
 106  FORMAT(A, INT_FMT, INT_FMT, INT_FMT, A, A, INT_FMT)
!===========================================================================
!
!  Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!  Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
!
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF
!
!  Get number of communicators
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
!
      FAIL = 0
      VERIFY = 0
      LOOP_COUNT = 0
!
!---------------------  Loop over Communicators  ------------------------
!
      DO 4000 CNUMBR = 1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(CNUMBR, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(CNUMBR, COMM_TYPE)
        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              TEST_NUMP)
!
!       Skip to end of loop if this node is not part of current
!       communicator
!
        IF (MPITEST_CURRENT_RANK .EQ. MPI_UNDEFINED ) GO TO 3800
!
!  Dup the communicator
!
!
        CALL MPI_COMM_DUP(COMM,  &
                          COMM2,  &
                          IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_DUP ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1

!
!  Compare the two communicators
!
        CALL MPI_COMM_COMPARE(COMM,  &
                              COMM2,  &
                              RESULT,  &
                              IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR, &
            ')  From:  MPI_COMM_COMPARE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF ( RESULT .NE. MPI_CONGRUENT ) THEN
          WRITE(INFOBUF,102) 'MPI_Comm_compare returned', RESULT,  &
            ',  expected', MPI_CONGRUENT,  &
            ' Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Check for inter-communicator
!
        CALL MPI_COMM_TEST_INTER(COMM,  &
                                 INTERCOMM,  &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_TEST_INTER #1 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPI_COMM_TEST_INTER(COMM2,  &
                                 INTERCOMM2,  &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_TEST_INTER #2 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF ( INTERCOMM .NEQV. INTERCOMM2 ) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_TEST_INTER returned', INTERCOMM2,  &
            ',  expected', INTERCOMM,  &
            ' Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Check the Size
!
        CALL MPI_COMM_SIZE(COMM2,  &
                           RESULTDUP,  &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_SIZE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        IF(INTERCOMM) THEN
!                             An inter-communicator TEST_NUMP will be too big
!
          CALL MPI_COMM_SIZE(COMM,  &
                             RESULT,  &
                             IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
              ')  From:  MPI_COMM_SIZE ( COMM_INDEX ',  &
              COMM_INDEX
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            FAIL = FAIL + 1
          END IF
        ELSE
          RESULT = TEST_NUMP
        END IF
!
        IF(RESULT .NE. RESULTDUP) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_SIZE returned', RESULTDUP,  &
            ',  expected', RESULT,  &
            ' Comm_Index = ', COMM_INDEX 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Check my rank 
!
        CALL MPI_COMM_RANK(COMM2, RESULTDUP, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Comm_rank ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF (RESULTDUP .NE. MPITEST_CURRENT_RANK) THEN
          WRITE(INFOBUF,102) 'MPI_Comm_rank returned', RESULTDUP,  &
            ',  expected', MPITEST_CURRENT_RANK,  &
            ' Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Now Isend 3 messages from rank 0 to rank 1, one per communicator.
!   Recv them in the opposite order, to make sure messages are
!   filtered to the proper communicator. To keep things simple
!   here, we'll test intercommunicators in a different test.
!   Repeat for Send 

        IF((TEST_NUMP .GE. 2)  .AND.  (.NOT. INTERCOMM )) THEN
          IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
!     Sender
            BUFFER1=1
            BUFFER2=2
            BUFFER3=3
!
            CALL MPI_ISEND(BUFFER1, 1, MPI_INTEGER, 1, 1, COMM,   &
                                                      REQUEST1, IERR)
            CALL MPI_ISEND(BUFFER2, 1, MPI_INTEGER, 1, 1, COMM2,  &
                                                      REQUEST2, IERR)
            CALL MPI_ISEND(BUFFER3, 1, MPI_INTEGER, 1, 1, COMM2, &
                                                      REQUEST3, IERR)
!
            CALL MPI_BARRIER(COMM2, IERR)
!
            CALL MPI_WAIT(REQUEST1, STATUS1, IERR)
            CALL MPI_WAIT(REQUEST2, STATUS2, IERR)
            CALL MPI_WAIT(REQUEST3, STATUS3, IERR)
!
            CALL MPI_SEND(BUFFER1, 1, MPI_INTEGER, 1, 1, COMM,  IERR)
            CALL MPI_SEND(BUFFER2, 1, MPI_INTEGER, 1, 1, COMM,  IERR)
            CALL MPI_SEND(BUFFER3, 1, MPI_INTEGER, 1, 1, COMM2, IERR)
!
          ELSE IF (MPITEST_CURRENT_RANK .EQ. 1) THEN
!     Receiver 
            BUFFER1=0
            BUFFER2=0
            BUFFER3=0
!
            CALL MPI_BARRIER(COMM2, IERR)
!
            CALL MPI_IRECV(BUFFER2, 1, MPI_INTEGER, 0, 1, COMM2,  &
                                                       REQUEST2, IERR)
            CALL MPI_IRECV(BUFFER3, 1, MPI_INTEGER, 0, 1, COMM2,  &
                                                       REQUEST3, IERR)
            CALL MPI_IRECV(BUFFER1, 1, MPI_INTEGER, 0, 1, COMM,   &
                                                       REQUEST1, IERR)
            CALL MPI_WAIT(REQUEST1, STATUS1, IERR)
            CALL MPI_WAIT(REQUEST2, STATUS2, IERR)
            CALL MPI_WAIT(REQUEST3, STATUS3, IERR)
!
            IF ((BUFFER1 .NE. 1) .OR.  (BUFFER2 .NE. 2) .OR.  &
                                        (BUFFER3 .NE.3)) THEN
              WRITE(INFOBUF,106)  &
                'Receive comms misfiltered messages (Sends):',  &
                BUFFER1, BUFFER2, BUFFER3, &
                ',  Expected:  1/2/3  ', &
                ' Comm_Index = ', COMM_INDEX 
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            END IF
            LOOP_COUNT = LOOP_COUNT + 1
!
            BUFFER1 = 0
            BUFFER2 = 0
            BUFFER3 = 0
!
            CALL MPI_IRECV(BUFFER1, 1, MPI_INTEGER, 0, 1,  &
                                              COMM,  REQUEST2, IERR)
            CALL MPI_IRECV(BUFFER3, 1, MPI_INTEGER, 0, 1,  &
                                              COMM2, REQUEST3, IERR)
            CALL MPI_IRECV(BUFFER2, 1, MPI_INTEGER, 0, 1, &
                                              COMM,  REQUEST1, IERR)
            CALL MPI_WAIT(REQUEST1, STATUS1, IERR)
            CALL MPI_WAIT(REQUEST2, STATUS2, IERR)
            CALL MPI_WAIT(REQUEST3, STATUS3, IERR)
!
            IF ((BUFFER1 .NE. 1) .OR.  (BUFFER2 .NE. 2) .OR.  &
                                        (BUFFER3 .NE.3)) THEN
              WRITE(INFOBUF,106)  &
                'Receive comms misfiltered messages (Sends):',  &
                BUFFER1, BUFFER2, BUFFER3, &
                ',  Expected:  1/2/3  ', &
                ' Comm_Index = ', COMM_INDEX 
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            END IF
            LOOP_COUNT = LOOP_COUNT + 1
          ELSE
            CALL MPI_BARRIER(COMM2, IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
                ')  From:  MPI_BARRIER ( COMM_INDEX ',  &
                  COMM_INDEX
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              ERRORSTRING = ' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
              WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            END IF
          END IF	
!
        END IF
!     Lastly, leave a communication outstanding, free the comm,
!     then complete the communication 
!
        IF ((TEST_NUMP .GE. 2)  .AND.   (.NOT.INTERCOMM )) THEN
          IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
!  Sender
            CALL MPI_ISEND(BUFFER1, 1, MPI_INTEGER, 1, 1, COMM2,  &
                                                       REQUEST1, IERR)
          ELSE IF (MPITEST_CURRENT_RANK .EQ. 1) THEN
!  Receiver
            BUFFER1 = 0
            CALL MPI_IRECV(BUFFER1, 1, MPI_INTEGER, 0, 1, COMM2,  &
                                                     REQUEST1, IERR)
          END IF
        END IF
!
!  Free the Dup'ed Communicator
!
        CALL MPI_COMM_FREE(COMM2, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_FREE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF( COMM2 .NE. MPI_COMM_NULL ) THEN
          WRITE(INFOBUF,103)  &
            'COMM2 = ', COMM2, '  after MPI_Comm_free, ',  &
            ',  expected  MPI_COMM_NULL',  &
            ' Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Now finish the communication started above 
        IF ((TEST_NUMP .GE.  2)  .AND.  (.NOT.INTERCOMM)) THEN
          IF (MPITEST_CURRENT_RANK .EQ.  0) THEN
!     Sender
            CALL MPI_WAIT(REQUEST1, STATUS1, IERR)
!
          ELSE IF (MPITEST_CURRENT_RANK .EQ. 1)  THEN
!     Receiver
            CALL MPI_WAIT(REQUEST1, STATUS1, IERR)
            IF (BUFFER1 .NE.  1) THEN
              WRITE(INFOBUF,102)  &
                'Receive while freeing comm received incorrect data', &
                BUFFER1, ',  expected  1,  Comm_Index = ',  &
                COMM_INDEX 
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            END IF
            LOOP_COUNT = LOOP_COUNT + 1
          END IF
        END IF
!
        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM)
 3800   CONTINUE
!                Node defined in current communicator
 4000 CONTINUE
!                End communicator loop
!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_COUNT - FAIL, FAIL, VERIFY, TESTNAME)
!      
      CALL MPI_FINALIZE(IERR)
!
      END
