! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
!  		  Test for MPI_Group_compare
!  
!  This test verifies that MPI_Group_compare compares correctly
!  against another group from an arbitrary communicator.  Each configured
!  communicator is tested.
!  
!  The following comparisons are made:
!   1 - group against group w/ all ranks reversed
!   2 - same but reversed
!   3 - group against itself.
!   4 - group against MPI_GROUP_EMPTY
!   5 - same but reversed
!   6 - group missing one rank
!   7 - same but reversed
!   
!  
!  As a consequence of this test  MPI_Group_incl and MPI_Group_free are also
!  tested here.
!  
!  Test history:
!     1  08/28/96     gt       Original version
! 
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'
!
      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
!  Program constants (Parameters)
!
      CHARACTER*32          TESTNAME
      PARAMETER( TESTNAME = 'MPI_GROUP_COMPARE')
!
!  General variables that occur in all programs
!
      INTEGER    CNUMBR
!                               index for communications loop
      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    FTEST_NUMP
!                               number of ranks in the current communicator
      INTEGER    LOOP_COUNT
!                               counts total number of failures, loops
      INTEGER    COMM
!                               MPITEST  communicator under test
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ERRORCLASS
      INTEGER    SIZE 
!                               error class of IERR
      INTEGER    ERRSIZE
!                               length of error message
      INTEGER    FAIL              
      INTEGER    VERIFY
!                           counts total number of PASSes/FAILures/Verifies
      CHARACTER*(IOSIZE)    INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
!  Variables specific to this program
!
      INTEGER    RANKS1(0:MAX_RANKS)
!                               array of ranks in group
      INTEGER    GROUP_SIZE
      INTEGER    RESULT
      INTEGER    EXPECT
      INTEGER    I
      INTEGER    GROUP1, GROUP2, GROUP3
!                               group handles
!-----------------------------------------------------------------------------
 98   FORMAT(A,A )
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT, A, INT_FMT)
 102  FORMAT(A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT)
 103  FORMAT(A, INT_FMT, A, A, A, INT_FMT)
 104  FORMAT(A, INT_FMT, A, A, INT_FMT)
!===========================================================================
!
!  Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!  Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
!
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF
!
!  Get number of communicators
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
!
      FAIL = 0
      VERIFY = 0
      LOOP_COUNT = 0
!
!---------------------  Loop over Communicators  ------------------------
!
      DO 4000 CNUMBR = 1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(CNUMBR, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(CNUMBR, COMM_TYPE)
        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              FTEST_NUMP)
!
!       Skip to end of loop if this node is not part of current
!       communicator
!
        IF (MPITEST_CURRENT_RANK .EQ. MPI_UNDEFINED ) GO TO 3800
!
! Make a Group
!
        CALL MPI_COMM_GROUP(COMM,  &
                            GROUP1,  &
                            IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_GROUP ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Get the size of the new group
!
        CALL MPI_GROUP_SIZE(GROUP1,  &
                            GROUP_SIZE,  &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_SIZE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Reverse the ranks
!
        DO 400 I = 0, GROUP_SIZE-1
          RANKS1 (I) = GROUP_SIZE -1 - I
 400    CONTINUE
!
        CALL MPI_GROUP_INCL(GROUP1, GROUP_SIZE, RANKS1, GROUP2, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_INCL #1 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Compare Groups #1
!
        CALL MPI_GROUP_COMPARE(GROUP1,  &
                               GROUP2, &
                               RESULT,  &
                               IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_COMPARE # 1 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the result
!
        EXPECT = MPI_SIMILAR
        IF(GROUP_SIZE .EQ. 1) EXPECT = MPI_IDENT
!
        IF(RESULT .NE. EXPECT) THEN
          WRITE(INFOBUF,102)  &
            'Group compare #1 returned (', RESULT, &
            ')  not equal to Expected (', EXPECT,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Compare Groups #2
!
        CALL MPI_GROUP_COMPARE(GROUP2,  &
                               GROUP1, &
                               RESULT,  &
                               IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_COMPARE #2 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the result
!
        IF(RESULT .NE. EXPECT) THEN
          WRITE(INFOBUF,102)  &
            'Group compare #2 returned (', RESULT, &
            ')  not equal to Expected (', EXPECT,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Compare #3, comparing a Group to itself
!
        CALL MPI_GROUP_COMPARE(GROUP1,  &
                               GROUP1, &
                               RESULT,  &
                               IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_COMPARE #3 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the result
!
        IF(RESULT .NE. MPI_IDENT) THEN
          WRITE(INFOBUF,104)  &
            'Group compare #3 returned (', RESULT, &
            ')  Expected (MPI_IDENT)',  &
            'COMM_INDEX = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Compare #4, comparing a Group to an empty group
!
        CALL MPI_GROUP_COMPARE(GROUP1,  &
                               MPI_GROUP_EMPTY, &
                               RESULT,  &
                               IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_COMPARE #4 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the result
!
        IF(RESULT .NE. MPI_UNEQUAL) THEN
          WRITE(INFOBUF,104)  &
            'Group compare #4 returned (', RESULT, &
            ')  Expectd (MPI_UNEQUAL)',  &
            'COMM_INDEX = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Compare #5, comparing an empty group to a group
!
        CALL MPI_GROUP_COMPARE(MPI_GROUP_EMPTY,  &
                               GROUP2, &
                               RESULT,  &
                               IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_COMPARE #5 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the result
!
        IF(RESULT .NE. MPI_UNEQUAL) THEN
          WRITE(INFOBUF,104)  &
            'Group compare #5 returned (', RESULT, &
            ')  Expectd (MPI_UNEQUAL)',  &
            'COMM_INDEX = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Compare #6, do a group less one rank
!
        CALL MPI_GROUP_INCL(GROUP1, GROUP_SIZE-1, RANKS1, GROUP3, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_INCL #2 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPI_GROUP_COMPARE(GROUP1,  &
                               GROUP3, &
                               RESULT,  &
                               IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_COMPARE #6 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the result
!
        IF(RESULT .NE. MPI_UNEQUAL) THEN
          WRITE(INFOBUF,104)  &
            'Group compare #6 returned (', RESULT, &
            ')  Expectd (MPI_UNEQUAL)',  &
            'COMM_INDEX = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Free the Groups
!
        CALL MPI_GROUP_FREE(GROUP1, &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_FREE for GROUP1 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPI_GROUP_FREE(GROUP2, &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_FREE for GROUP2 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPI_GROUP_FREE(GROUP3, &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_FREE for GROUP3 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM)
 3800   CONTINUE
!                Node defined in current communicator
 4000 CONTINUE
!                End communicator loop
!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_COUNT - FAIL, FAIL, 0, TESTNAME)
!      
      CALL MPI_FINALIZE(IERR)
!
      END
