! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
!  		  Test for MPI_Group_translate_ranks
!  
!  This test verifies that MPI_Group_translate_ranks translates correctly 
!  against another group from an arbitrary commuicator.  Each configured
!  communicator is tested.
!  
!  The following comparisons are made:
!   1 - group against group w/ all ranks reversed
!   2 - group w/ all ranks reversed against group
!   3 - same as #2, but only one rank
!   4 - same as #1, but only one rank
!   5 - self against group w/ all ranks reversed
!   6 - group against MPI_GROUP_EMPTY 
!   7 - group against self (all ranks)
!   8 - group against itself.
!  
!  As a consequence of this test, MPI_Group_size, MPI_Group_rank,
!  MPI_Group_incl, and MPI_Group_free are also tested here.
!  
!  Test history:
!     1  08/28/96     gt       Original version
! 
!***********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'
!
      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
!  Program constants (Parameters)
!
      CHARACTER*32          TESTNAME
      PARAMETER( TESTNAME = 'MPI_GROUP_TRANSLATE_RANKS')
!
!  General variables that occur in all programs
!
      INTEGER    CNUMBR
!                               index for communications loop
      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    FTEST_NUMP
!                               number of ranks in the current communicator
      INTEGER    LOOP_COUNT
!                               counts total number of failures, loops
      INTEGER    COMM
!                               MPITEST  communicator under test
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ERRORCLASS
      INTEGER    SIZE 
!                               error class of IERR
      INTEGER    ERRSIZE
!                               length of error message
      INTEGER    FAIL              
      INTEGER    VERIFY
!                           counts total number of PASSes/FAILures/Verifies
      CHARACTER*(IOSIZE)    INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
!
!  Variables specific to this program
!
      INTEGER    RANKS1(0:MAX_RANKS)
      INTEGER    RANKS2(0:MAX_RANKS)
!                               array of ranks in group
      INTEGER    RESULT
      INTEGER    EXPECT
      INTEGER    I
!
      INTEGER    COMM_SIZE
      INTEGER    GROUP_SIZE
!
      INTEGER    GROUP
      INTEGER    GROUP2
      INTEGER    GROUPSELF
!                               group handles
!-----------------------------------------------------------------------------
 98   FORMAT(A,A )
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT, A, INT_FMT)
 102  FORMAT(A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT)
 103  FORMAT(A, INT_FMT, A, A, A, INT_FMT)
 104  FORMAT(A, INT_FMT, A, A, INT_FMT)
!===========================================================================
!
!  Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!  Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
!
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF
!
!  Get number of communicators
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)

      FAIL = 0
      VERIFY = 0
      LOOP_COUNT = 0
!
!---------------------  Loop over Communicators  ------------------------
!
      DO 4000 CNUMBR = 1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(CNUMBR, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(CNUMBR, COMM_TYPE)
        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              FTEST_NUMP)
!
!       Skip to end of loop if this node is not part of current
!       communicator
!
        IF (MPITEST_CURRENT_RANK .EQ. MPI_UNDEFINED ) GO TO 3800
!
!   Get the size of the current communicator
!
        CALL MPI_COMM_SIZE(COMM,  &
                           COMM_SIZE,  &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_SIZE #1 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!-------------------------------  MPI_Comm_group  ---------------------------
!
        CALL MPI_COMM_GROUP(COMM,  &
                            GROUP,  &
                            IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_GROUP #1 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Get the size of the new group
!
        CALL MPI_GROUP_SIZE(GROUP,  &
                            GROUP_SIZE,  &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_SIZE #1 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Reverse the ranks
!
        DO 400 I = 0, GROUP_SIZE-1
          RANKS1 (I) = GROUP_SIZE -1 - I
 400    CONTINUE
!
        CALL MPI_GROUP_INCL(GROUP,  &
                            GROUP_SIZE,  &
                            RANKS1,  &
                            GROUP2,  &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_INCL #1 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!   Check the size of the new group
!
        CALL MPI_GROUP_SIZE(GROUP2,  &
                            RESULT,  &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_SIZE #2 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        IF(RESULT .NE. GROUP_SIZE) THEN
          WRITE(INFOBUF,102)  &
            'MPI_Group_size #2 returned (', RESULT, &
            ')  not equal to Expected (', EXPECT,  &
            ')   (COMM_INDEX = ', COMM_INDEX, ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check my rank in the new group
!
        CALL MPI_GROUP_RANK(GROUP2, &
                            RESULT, &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_RANK '
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!
        IF(RESULT .NE. GROUP_SIZE - 1 - MPITEST_CURRENT_RANK) THEN
          WRITE(INFOBUF,102)  &
            'MPI_Group_rank returned (', RESULT, &
            ')   (COMM_INDEX = ', COMM_INDEX, ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Translate all the new ranks
!
        DO 520 I = 0,  GROUP_SIZE-1
          RANKS1(I) = I
 520    CONTINUE
!
!-----------------------  MPI_Group_translate_ranks  -----------------------
!
        CALL MPI_GROUP_TRANSLATE_RANKS(GROUP,  &
                                       GROUP_SIZE, &
                                       RANKS1, &
                                       GROUP2, &
                                       RANKS2, &
                                       IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_TRANSLATE_RANKS # 1 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the translation
!
        VERIFY = 0
        DO 580 I = 0, GROUP_SIZE-1
          IF(RANKS2(I) .NE. GROUP_SIZE - 1 - I) VERIFY = VERIFY + 1
 580    CONTINUE
!
        IF(VERIFY .NE. 0) THEN
          WRITE(INFOBUF,102)  &
            'Translate_ranks #1 Errors = (', VERIFY, &
            ')    (COMM_INCEX =',         COMM_INDEX,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Translate IN REVERSE
!
!
!-----------------------  MPI_Group_translate_ranks  -----------------------
!
        CALL MPI_GROUP_TRANSLATE_RANKS(GROUP2,  &
                                       GROUP_SIZE, &
                                       RANKS1, &
                                       GROUP, &
                                       RANKS2, &
                                       IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_TRANSLATE_RANKS #2 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the translation
!
        VERIFY = 0
        DO 640 I = 0, GROUP_SIZE-1
          IF(RANKS2(I) .NE. GROUP_SIZE - 1 - I) VERIFY = VERIFY + 1
 640    CONTINUE
!
        IF(VERIFY .NE. 0) THEN
          WRITE(INFOBUF,102)  &
            'Translate_ranks #2 Errors = (', VERIFY, &
            ')    (COMM_INCEX =',         COMM_INDEX,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Just Translate my own rank
!
        RANKS1(0) = MPITEST_CURRENT_RANK
!
!-----------------------  MPI_Group_translate_ranks  -----------------------
!
        CALL MPI_GROUP_TRANSLATE_RANKS(GROUP2,  &
                                       1, &
                                       RANKS1, &
                                       GROUP, &
                                       RANKS2, &
                                       IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_TRANSLATE_RANKS #3 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the translation
!
        EXPECT = GROUP_SIZE - 1 - MPITEST_CURRENT_RANK
        IF(RANKS2(0) .NE. EXPECT) THEN
          WRITE(INFOBUF,102)  &
            'Translate_ranks #3 Error:  Actual = (', RANKS2(0), &
            ')    Expected = (', EXPECT, &
            ')    (COMM_INCEX =',         COMM_INDEX,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Translate against MPI_COMM_SELF
!
!
!-------------------------------  MPI_Comm_group  ---------------------------
!
        CALL MPI_COMM_GROUP(MPI_COMM_SELF,  &
                            GROUPSELF,  &
                            IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_GROUP(SELF) ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1

!
        RANKS1(0) = MPITEST_CURRENT_RANK
!
!-----------------------  MPI_Group_translate_ranks  -----------------------
!
        CALL MPI_GROUP_TRANSLATE_RANKS(GROUP,  &
                                       1, &
                                       RANKS1, &
                                       GROUPSELF, &
                                       RANKS2, &
                                       IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_TRANSLATE_RANKS #4 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the translation
!
        EXPECT = 0
        IF(RANKS2(0) .NE. EXPECT) THEN
          WRITE(INFOBUF,102)  &
            'Translate_ranks(SELF) #4 Error:  Actual = (', RANKS2(0), &
            ')    Expected = (', EXPECT, &
            ')    (COMM_INCEX =',         COMM_INDEX,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Translate against MPI_COMM_SELF in Reverse
!
!
        RANKS1(0) = 0
!
!-----------------------  MPI_Group_translate_ranks  -----------------------
!
        CALL MPI_GROUP_TRANSLATE_RANKS(GROUPSELF,  &
                                       1, &
                                       RANKS1, &
                                       GROUP2, &
                                       RANKS2, &
                                       IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_TRANSLATE_RANKS #5 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the translation
!
        EXPECT = GROUP_SIZE - 1 - MPITEST_CURRENT_RANK
        IF(RANKS2(0) .NE. EXPECT) THEN
          WRITE(INFOBUF,102)  &
            'Translate_ranks(SELF) #4 Error:  Actual = (', RANKS2(0), &
            ')    Expected = (', EXPECT, &
            ')    (COMM_INCEX =',         COMM_INDEX,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Translate against MPI_GROUP_EMPTY   -- all should be undefined
!
        
        DO 720 I = 0, GROUP_SIZE-1
          RANKS1(I) = I
 720    CONTINUE
!
!-----------------------  MPI_Group_translate_ranks  -----------------------
!
        CALL MPI_GROUP_TRANSLATE_RANKS(GROUP,  &
                                       GROUP_SIZE, &
                                       RANKS1, &
                                       MPI_GROUP_EMPTY, &
                                       RANKS2, &
                                       IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_TRANSLATE_RANKS #6 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the translation
!
        VERIFY = 0
        DO 740 I = 0, GROUP_SIZE-1
          IF(RANKS2(I) .NE. MPI_UNDEFINED) VERIFY = VERIFY + 1
 740    CONTINUE
!
        IF(VERIFY .NE. 0) THEN
          WRITE(INFOBUF,102)  &
            'Translate_ranks #6 Errors = (', VERIFY, &
            ')    (COMM_INCEX =',         COMM_INDEX,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Translate against MPI_COMM_SELF   -- all UNDEFINED, but my rank
!
        DO 820 I = 0, GROUP_SIZE-1
          RANKS1(I) = I
 820    CONTINUE
!
!-----------------------  MPI_Group_translate_ranks  -----------------------
!
        CALL MPI_GROUP_TRANSLATE_RANKS(GROUP,  &
                                       GROUP_SIZE, &
                                       RANKS1, &
                                       GROUPSELF, &
                                       RANKS2, &
                                       IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_TRANSLATE_RANKS #7 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the translation
!
        VERIFY = 0
        DO 840 I = 0, GROUP_SIZE-1
          IF( I .NE. MPITEST_CURRENT_RANK  .AND. &
              RANKS2(I) .NE. MPI_UNDEFINED) THEN
            VERIFY = VERIFY + 1
          ELSE IF( I .EQ. MPITEST_CURRENT_RANK  .AND. &
                  RANKS2(I) .NE. 0) THEN
            VERIFY = VERIFY + 1
          END IF            
 840    CONTINUE
!
        IF(VERIFY .NE. 0) THEN
          WRITE(INFOBUF,102)  &
            'Translate_ranks #7 Errors = (', VERIFY, &
            ')    (COMM_INCEX =',         COMM_INDEX,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!------------------------------------------------------------------------------
!------------------------------------------------------------------------------
!
!     Finally, the group against itself
!
!
        DO 920 I = 0, GROUP_SIZE-1
          RANKS1(I) = group_size - I - 1
 920    CONTINUE
!
!-----------------------  MPI_Group_translate_ranks  -----------------------
!
        CALL MPI_GROUP_TRANSLATE_RANKS(GROUP,  &
                                       GROUP_SIZE, &
                                       RANKS1, &
                                       GROUP, &
                                       RANKS2, &
                                       IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_TRANSLATE_RANKS #8 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the translation
!
        VERIFY = 0
        DO 940 I = 0, GROUP_SIZE-1
          IF(RANKS2(I) .NE. RANKS1(I)) VERIFY = VERIFY + 1
 940    CONTINUE
!
        IF(VERIFY .NE. 0) THEN
          WRITE(INFOBUF,102)  &
            'Translate_ranks #8 Errors = (', VERIFY, &
            ')    (COMM_INCEX =',         COMM_INDEX,  &
            ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF

!------------------------------------------------------------------------------
!------------------------------------------------------------------------------
!
!  Free the Groups
!
        CALL MPI_GROUP_FREE(GROUPSELF, &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_FREE #1 for GROUPSELF ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPI_GROUP_FREE(GROUP2, &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_FREE #2 for GROUP2 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPI_GROUP_FREE(GROUP, &
                            IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_GROUP_FREE #3 for GROUP ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM)
 3800   CONTINUE
!                Node defined in current communicator
 4000 CONTINUE
!                End communicator loop
!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_COUNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END
