! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
!
! Copyright - 1996 Intel Corporation
!
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
!
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
!
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
!
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
!
! ----------------------------------------------------------------------
! **********************************************************************
!                   Test for MPI_Bsend_init()
!                   and MPI_Startall()
!
! This test verifies that the basic non-blocking MPI_Bsend_init operation with
! MPI_Recv_init delivers messages in the order they were sent.
!
! This test uses the first 2 ranks in MPI_COMM_WORLD, first sending a large
! message then a number of short ones, and ensures that they are received
! in the proper order by verifying the data received.
!
! **********************************************************************
#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    IERR
      INTEGER    ERR
      INTEGER    SIZE
!                               return values from MPI calls
      INTEGER    I
      INTEGER    ERRORS

      INTEGER    RECV_STAT(MPI_STATUS_SIZE)

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
       CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

      INTEGER NUMSHORT
      PARAMETER (NUMSHORT=20)

      INTEGER LONGLENGTH
      PARAMETER (LONGLENGTH=8192)

      INTEGER BUFFLENGTH, BL
      PARAMETER (BUFFLENGTH= &
               2*LONGLENGTH + (NUMSHORT+1)*(MPI_BSEND_OVERHEAD*2))

      INTEGER    STAT(MPI_STATUS_SIZE, NUMSHORT+1)
      INTEGER    REQUEST(NUMSHORT+1)

      MPITEST_BUF_TYPE VALUE

      CHARACTER*(NUMSHORT+1) SHORTB
      CHARACTER*(LONGLENGTH+1) BUFFER
      CHARACTER*(BUFFLENGTH) BSEND_BUFF

 99   FORMAT(A,INT_FMT)
 189  FORMAT(A,INT_FMT,A)

!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Startall1'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF


!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0

!
!     Need at least 2 nodes.
!
      IF (MPITEST_NUMP .LT. 2) THEN
        INFOBUF = 'At least 2 ranks required to run this test'
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF


      IF (MPITEST_ME .LT. 2) THEN
        IF (MPITEST_ME .EQ. 0) THEN
!
!         Sender
!
          CALL MPI_BUFFER_ATTACH(BSEND_BUFF, BUFFLENGTH, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_BUFFER_ATTACH() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF


          CALL MPITEST_DATATEMPLATE_INIT(VALUE, 0, MPITEST_CHARACTER)
          CALL MPITEST_INIT_BUFFER_INC(MPITEST_CHARACTER, LONGLENGTH, &
                VALUE, BUFFER)
          CALL MPITEST_DATATEMPLATE_INIT(VALUE, 1, MPITEST_CHARACTER)
          CALL MPITEST_INIT_BUFFER_INC(MPITEST_CHARACTER, NUMSHORT, &
                VALUE, SHORTB)

          CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)

!
!          Send a long message
!
          CALL MPI_BSEND_INIT(BUFFER, LONGLENGTH, MPI_CHARACTER, 1, 1, &
                 MPI_COMM_WORLD, REQUEST(1), IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_BSEND_INIT(1) returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF

!
!         Send a number of short messages
!
          DO 200 I=1, NUMSHORT
            CALL MPI_BSEND_INIT(SHORTB(I:I), 1, MPI_CHARACTER, 1, 1, &
                 MPI_COMM_WORLD, REQUEST(I+1), IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              WRITE(INFOBUF,99) 'MPI_BSEND_INIT(2) returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
            END IF
 200      CONTINUE

          CALL MPI_STARTALL(NUMSHORT+1, REQUEST, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_STARTALL() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF

          CALL MPI_WAITALL(NUMSHORT+1, REQUEST, STAT, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_WAITALL() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF

          LOOP_CNT = 1 + NUMSHORT

!         The second parameter is an OUTPUT value, so you must not 
!         use a parameter value
          CALL MPI_BUFFER_DETACH(BSEND_BUFF, BL, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_BUFFER_DETACH() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF


        ELSE
!
!         Receiver
!
          CALL MPITEST_DATATEMPLATE_INIT(VALUE, -1, MPITEST_CHARACTER)
          CALL MPITEST_INIT_BUFFER(MPITEST_CHARACTER, LONGLENGTH, &
                VALUE, BUFFER)
          CALL MPITEST_INIT_BUFFER(MPITEST_CHARACTER, NUMSHORT, &
                VALUE, SHORTB)


          CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)

!
!         Receive the long message
!
          CALL MPI_RECV_INIT(BUFFER, LONGLENGTH, MPI_CHARACTER, 0, 1, &
               MPI_COMM_WORLD, REQUEST(1), IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_RECV_INIT(1) returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF

!
!         Receive a number of short messages
!
          DO 300 I=1, NUMSHORT
            CALL MPI_RECV_INIT(SHORTB(I:I), 1, MPI_CHARACTER, 0, &
                 1, MPI_COMM_WORLD, REQUEST(I+1), IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              WRITE(INFOBUF,99) 'MPI_RECV_INIT(2) returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
            END IF
 300      CONTINUE

          CALL MPI_STARTALL(NUMSHORT+1, REQUEST, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_STARTALL() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF

          CALL MPI_WAITALL(NUMSHORT+1, REQUEST, STAT, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_WAITALL() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF


          LOOP_CNT = 1 + NUMSHORT

!
!         Check received data
!
          CALL MPITEST_DATATEMPLATE_INIT(VALUE, 0, MPITEST_CHARACTER)
          CALL MPITEST_BUFFER_ERRORS_INC(MPITEST_CHARACTER, LONGLENGTH, &
               VALUE, BUFFER, ERRORS)

          IF (ERRORS .NE. 0) THEN
            FAIL = FAIL + 1
            INFOBUF=' '
            WRITE (INFOBUF,189) 'Found', ERRORS, &
               ' errors in first (long) buffer '
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          END IF

          CALL MPITEST_DATATEMPLATE_INIT(VALUE, 1, MPITEST_CHARACTER)
          CALL MPITEST_BUFFER_ERRORS_INC(MPITEST_CHARACTER, NUMSHORT, &
               VALUE, SHORTB, ERRORS)

          IF (ERRORS .NE. 0) THEN
            FAIL = FAIL + ERRORS
            INFOBUF=' '
            WRITE (INFOBUF,189) 'Found', ERRORS, &
               ' short messages received incorrectly '
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          END IF

        END IF

      DO 400 I=1, NUMSHORT+1
        CALL MPI_REQUEST_FREE(REQUEST(I), IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,99) 'MPI_REQUEST_FREE() returned', IERR
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          INFOBUF=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
        END IF
 400  CONTINUE

      ELSE
!
!       Ranks >= 2 need to match BARRIER above
!
        CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)

      END IF


!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END

