! -----------------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose
! and without fee, provided that the above copyright notice and the following
! paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases comprising
! this suite are correct or are an accurate representation of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT OR
! SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING, CONSEQUENTIAL,
! INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT LIMITED TO INFRINGEMENT,
! LOSS OF USE, BUSINESS INTERRUPTIONS, AND LOSS OF PROFITS, IRRESPECTIVE OF
! WHETHER INTEL HAS ADVANCE NOTICE OF THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING, BUT NOT
! LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
! PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED HEREUNDER
! IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO OBLIGATION TO PROVIDE
! MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR MODIFICATIONS.
! -----------------------------------------------------------------------------
! *****************************************************************************
!                           Test for MPI_Iprobe()
! Each rank (except the root rank) will do a send to the root rank.  The root
! rank will then do a MPI_Iprobe() followed by a MPI_Recv().  The buffer
! received may or may not be send and / or checked (depending on if
! MPITEST_BUFFER_RECV and / or MPITEST_BUFFER_CHK is / are #defined or not in
! this source).  The status object returned from MPI_Iprobe() may or may not
! be checked depending on if MPITEST_STATUS_CHK is defined or not.
! 
! This test may be run in any communicator with a minimum of 2 group members,
! with any data type, and with any non-negative message length.
! 
! The MPITEST environment provides looping over communicator size and
! message length (if MPITEST_STATUS_CHK is #defined)  The properties of the
! loops are encoded in configuration arrays in the file config.h .
! 
! MPI Calls dependencies for this test:
!   MPI_ISEND(), MPI_RECV(), MPI_CANCEL(), MPI_IPROBE(), MPI_INIT(), MPI_FINALIZE()
!   MPI_COMM_TEST_INTER(), MPI_BARRIER(), MPI_ERROR_STRING(), 
!   [MPI_GET_COUNT(), MPI_ALLREDUCE(), MPI_COMM_RANK(), MPI_COMM_SIZE()]
!  
!
! Test history:
!
!   1  06/20/96     simont       Original version
!
! *****************************************************************************
!
#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


!     The index of the current buffer type
      INTEGER TEST_TYPE

!     The length of the current buffer
      INTEGER LENGTH

!     The length of the current buffer in bytes
      INTEGER BYTE_LENGTH

!     The number of process in the current communicator
      INTEGER TEST_NUMP

!     The array index of the current communicator
      INTEGER COMM_INDEX

!     The index of the current communicator type
      INTEGER COMM_TYPE

!     Loop counter for the message length loop
      INTEGER LENGTH_COUNT

!     Loop counter for communicator loop
      INTEGER COMM_COUNT

!     The maximum of input communicator from config file 
      INTEGER MAX_COMM_COUNT

!     Communicator
      INTEGER COMM

!     Errors from one MPI call
      INTEGER ERROR

!     Counts total number of failure(s)
      INTEGER FAIL

!     Return size from MPI_Error_string
      INTEGER SIZE

!     Counts total number of loops through test
      INTEGER LOOP_CNT

!     Return value from MPI call
      INTEGER IERR, IERR2

!     Maximum bufer length specified in config file
      INTEGER MAX_LENGTH

!     Maximum buffer length in bytes
      INTEGER MAX_BYTE_LENGTH

!     The root rank
      INTEGER ROOT

!     The source rank
      INTEGER SOURCE

!     Message tag
      INTEGER TAG

!     Status object returned from MPI call
      INTEGER STATUS(MPI_STATUS_SIZE)

!     Request handle
      INTEGER REQUEST

!     Maximum number of rank participating in test
      INTEGER MAXNP
      INTEGER I, J

!     Logical flag
      LOGICAL FLAG

      INTEGER MPITEST_TAG 
      PARAMETER (MPITEST_TAG = 1)

!  Maximum number of processes to participate in message sending   
!  This is here so that the test can be run in arbitray number     
!  of MPI processess without potentially exhausting message buffer 
!  as the number of mpi processes created increase.                 
      INTEGER MPITEST_MAX_NP 
      PARAMETER (MPITEST_MAX_NP = 10)

#ifdef MPITEST_STATUS_CHK
      INTEGER COUNT, MAX_LENGTH_COUNT
#endif

      MPITEST_BUF_TYPE VALUE

      MPITEST_BUF_TYPE BUFFER(MAX_BUFF_SIZE+1)

      CHARACTER*(IOSIZE)  INFOBUF

      CHARACTER*32   TESTNAME
       CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

 10   FORMAT(A,INT_FMT)
 20   FORMAT(A)
 30   FORMAT(A, A, INT_FMT, A)
 40   FORMAT(A, INT_FMT, A)
 50   FORMAT(A, INT_FMT, A, INT_FMT, A, INT_FMT, A, INT_FMT, &
             A, INT_FMT, A, INT_FMT, A, INT_FMT)
 60   FORMAT(A, INT_FMT, A, INT_FMT)
 70   FORMAT(A, A, A, INT_FMT, A, INT_FMT)
 80   FORMAT(INT_FMT, A)


!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)

      IF (IERR .NE. MPI_SUCCESS) THEN
         WRITE(INFOBUF,10) 'MPI_INIT() returned ', IERR 
         CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
         INFOBUF=' '
         CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, IERR2)
         CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
      END IF

      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Iprobe_source'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      FAIL = 0
      LOOP_CNT = 0

!
!     Get number of communicators, data types and message lengths to test
!
      CALL MPITEST_NUM_COMM_SIZES(MAX_COMM_COUNT)

!
!     Get max message length for test
!
      CALL MPITEST_GET_MAX_MESSAGE_LENGTH(MAX_BYTE_LENGTH)

!
!     Loop through communicators
!
      DO 100 COMM_COUNT=1, MAX_COMM_COUNT, 1
         CALL MPITEST_GET_COMM_INDEX(COMM_COUNT, COMM_INDEX)
         CALL MPITEST_GET_COMM_TYPE(COMM_COUNT, COMM_TYPE)

         CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, &
                                       COMM, TEST_NUMP)

!        A measure to prevent having message flood the message buffer
!        when test scales up, all ranks larger than MPITEST_MAX_NP
!        would not participate in message sending.  MPITEST_MAX_NP
!        can be customized if needed at the top of this source
!
         IF (TEST_NUMP .GT. MPITEST_MAX_NP) THEN
            MAXNP = MPITEST_MAX_NP
         ELSE
            MAXNP = TEST_NUMP
         END IF

         IF (TEST_NUMP .LT. 2) THEN
!           Skipping communicator with comm size < 2
!
            WRITE(INFOBUF, 30) 'Skipping communicator with size < 2', &
                 ' (commtype: ', COMM_TYPE, ') for this test'
            CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
         ELSE
!           All other process(es) not in the current communicator
!           will skip this.
            IF (COMM .EQ. MPI_COMM_NULL) GO TO 800

            CALL MPI_COMM_TEST_INTER(COMM, FLAG, IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
               WRITE(INFOBUF, 10) &
                    'MPITEST_COMM_TEST_INTER() returned ', IERR
               CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
               INFOBUF=' '
               CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, IERR2)
               CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
            ELSE IF (FLAG .EQV. .TRUE.) THEN
               WRITE(INFOBUF, 40) &
                    'Skipping inter communicator (commtype: ', &
                    COMM_TYPE, ') for this test'
               CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
            ELSE
               TEST_TYPE = MPITEST_INTEGER

!              Convert the number of bytes in the maximum length
!              message into the number of elements of the current
!              type
!
               CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, &
                                            MAX_BYTE_LENGTH, &
                                            MAX_LENGTH)

#ifdef MPITEST_STATUS_CHK
               CALL MPITEST_NUM_MESSAGE_LENGTHS(MAX_LENGTH_COUNT)

               DO 200 LENGTH_COUNT=1, MAX_LENGTH_COUNT, 1
                  CALL MPITEST_GET_MESSAGE_LENGTH(LENGTH_COUNT, &
                                                  BYTE_LENGTH)
                  CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, &
                                               BYTE_LENGTH, &
                                               LENGTH)
#else
                  LENGTH_COUNT = 1
                  LENGTH = 1
#endif
                  DO 300 ROOT=0, MAXNP - 1, 1
!                    Print an informational message
!
                     IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
                        WRITE(INFOBUF, 50) '(', LENGTH_COUNT, ', ',  &
                             COMM_COUNT, ') length ', LENGTH, &
                             ' commsize ', TEST_NUMP, ' commtype ', &
                             COMM_TYPE, ' data type ', TEST_TYPE, &
                             ' root ', ROOT
                        CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
                     END IF

                     IF (MPITEST_CURRENT_RANK .NE. MPI_UNDEFINED) THEN
!                       Initialize send buffer
!
                        CALL MPITEST_DATATEMPLATE_INIT(VALUE, &
                                MPITEST_CURRENT_RANK, &
                                TEST_TYPE)
                        CALL MPITEST_INIT_BUFFER(TEST_TYPE, LENGTH, &
                                VALUE, BUFFER)

                        LOOP_CNT = LOOP_CNT + 1

                        TAG = MPITEST_TAG

                        IF (MPITEST_CURRENT_RANK .NE. ROOT .AND. &
                            MPITEST_CURRENT_RANK .LT. MAXNP) THEN
                           WRITE(INFOBUF, 60) 'Sending message to ', &
                                ROOT, ' tag: ', TAG
                           CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)

                           CALL MPI_ISEND(BUFFER, LENGTH, &
                                   MPITEST_MPI_DATATYPES(TEST_TYPE), &
                                   ROOT, TAG, COMM, REQUEST, IERR)
                           IF (IERR .NE. MPI_SUCCESS) THEN
                              FAIL = FAIL + 1
                              WRITE(INFOBUF, 10)  &
                                   'MPI_ISEND() returned ', IERR
                              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                                                   INFOBUF)
                              INFOBUF=' '
                              CALL MPI_ERROR_STRING(IERR, INFOBUF, &
                                                    SIZE, IERR2)
                              CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                                                   INFOBUF)
                           END IF
                        ELSE IF (MPITEST_CURRENT_RANK .EQ. ROOT) THEN
!                          Call MPI_IPROBE() with MPI_ANY_SOURCE first
!
                           WRITE(INFOBUF, 10) &
                              'Probing source MPI_ANY_SOURCE tag: ', &
                              TAG
                           CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                                   INFOBUF)

 111                       CALL MPI_IPROBE(MPI_ANY_SOURCE, TAG, &
                                          COMM, FLAG, STATUS, IERR)

                           IF (FLAG .NEQV. .TRUE.) GO TO 111

                           IF (IERR .NE. MPI_SUCCESS) THEN
                              FAIL = FAIL + 1
                              WRITE(INFOBUF, 10) &
                                   'MPI_IPROBE() returned ', IERR
                              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                                                   INFOBUF)
                              INFOBUF=' '
                              CALL MPI_ERROR_STRING(IERR, INFOBUF, &
                                                    SIZE, IERR2)
                              CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                                                   INFOBUF)
                           END IF

#ifdef MPITEST_STATUS_CHK
!                          Check status(MPI_SOURCE)
!
                           WRITE(INFOBUF, 20) &
                                'Verifying output status object'
                           CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                                                INFOBUF)

                           IF (STATUS(MPI_SOURCE) .LT. 0) THEN
                              FAIL = FAIL + 1
                              WRITE(INFOBUF, 30) &
                                 'STATUS(MPI_SOURCE) from ', &
                                 'MPI_IPROBE() is negative (', &
                                 STATUS(MPI_SOURCE), ')'
                              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                                      INFOBUF)
                           ELSE IF (STATUS(MPI_SOURCE) .GE. &
                                    TEST_NUMP) THEN
                              FAIL = FAIL + 1
                              WRITE(INFOBUF, 30) 'STATUS(MPI_SOURCE)', &
                                   ' from MPI_IPROBE() is too large (', &
                                   STATUS(MPI_SOURCE), ')'
                              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                                      INFOBUF)
                           END IF

!                          Check STATUS(MPI_TAG)
!
                           IF (STATUS(MPI_TAG) .NE. TAG) THEN
                              FAIL = FAIL + 1
                              WRITE(INFOBUF, 70) 'STATUS(MPI_TAG)', &
                                   ' from MPI_IPROBE() is unexpected', &
                                   ' Expected: ', TAG, ' Actual: ', &
                                   STATUS(MPI_TAG)
                              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                                      INFOBUF) 
                           END IF

!                          Not checkin STATUS(MPI_ERROR)
!

!                          Checking length of message to be received
!

                           CALL MPI_GET_COUNT(STATUS, &
                                   MPITEST_MPI_DATATYPES(TEST_TYPE), &
                                   COUNT, IERR)
                           IF (IERR .NE. MPI_SUCCESS) THEN
                              FAIL = FAIL + 1
                              WRITE(INFOBUF, 10) &
                                   'MPI_GET_COUNT() returned ', IERR
                              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                                                   INFOBUF)
                              INFOBUF=' '
                              CALL MPI_ERROR_STRING(IERR, INFOBUF, &
                                                    SIZE, IERR2)
                              CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                                                   INFOBUF)
                           END IF

                           IF (COUNT .NE. LENGTH) THEN
                              FAIL = FAIL + 1
                              WRITE(INFOBUF, 70) 'STATUS from ', &
                                   'MPI_IPROBE() has unexpected length', &
                                   ' Expected: ', LENGTH, 'Actual: ', &
                                   COUNT
                              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                                      INFOBUF)
                           END IF
#else
                           WRITE(INFOBUF, 20) &
                                'Output STATUS *not* verified'
                           CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                                   INFOBUF)
#endif

                           DO 400 SOURCE=0, MAXNP - 1, 1
                              IF (SOURCE .NE. ROOT) THEN
                                 WRITE(INFOBUF, 60) &
                                    'Probing source ', SOURCE, &
                                    ' tag: ', TAG
                                 CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                                         INFOBUF)

 222                             CALL MPI_IPROBE(SOURCE, TAG, COMM, &
                                         FLAG, STATUS, IERR)

                                 IF (FLAG .NEQV. .TRUE.) GO TO 222

                                 IF (IERR .NE. MPI_SUCCESS) THEN
                                    FAIL = FAIL + 1
                                    WRITE(INFOBUF, 10)  &
                                      'MPI_GET_COUNT() returned ', &
                                      IERR
                                    CALL MPITEST_MESSAGE( &
                                            MPITEST_NONFATAL, &
                                            INFOBUF)
                                    INFOBUF=' '
                                    CALL MPI_ERROR_STRING(IERR, &
                                            INFOBUF, &
                                            SIZE, IERR2)

                                    CALL MPITEST_MESSAGE( &
                                            MPITEST_FATAL, &
                                            INFOBUF)
                                 END IF


#ifdef MPITEST_STATUS_CHK
!                                Check status(MPI_SOURCE)
!
                                 WRITE(INFOBUF, 20) &
                                   'Verifying output status object'
                                 CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                                   INFOBUF)

                                 IF (STATUS(MPI_SOURCE) .NE.  &
                                     SOURCE) THEN
                                    FAIL = FAIL + 1
                                    WRITE(INFOBUF, 70) &
                                       'STATUS(MPI_SOURCE) from ', &
                                       'MPI_IPROBE() is unexpected', &
                                       ' Expected: ', SOURCE, &
                                       'Actual: ', &
                                       STATUS(MPI_SOURCE)
                                    CALL MPITEST_MESSAGE( &
                                       MPITEST_NONFATAL, &
                                       INFOBUF)
                                 END IF

!                                Check STATUS(MPI_TAG)
!
                                 IF (STATUS(MPI_TAG) .NE. TAG) THEN
                                    FAIL = FAIL + 1
                                    WRITE(INFOBUF, 70) &
                                       'STATUS(MPI_TAG) from ', &
                                       'MPI_IPROBE() is unexpected', &
                                       ' Expected: ', TAG, ' Actual: ', &
                                       STATUS(MPI_TAG)
                                    CALL MPITEST_MESSAGE( &
                                       MPITEST_NONFATAL, &
                                       INFOBUF)
                                 END IF

!                                Not checkin STATUS(MPI_ERROR)
!

!                                Checking length of message to be
!                                received
!

                                 CALL MPI_GET_COUNT(STATUS, &
                                    MPITEST_MPI_DATATYPES(TEST_TYPE), &
                                    COUNT, IERR)
                                 IF (IERR .NE. MPI_SUCCESS) THEN
                                    FAIL = FAIL + 1
                                    WRITE(INFOBUF, 10) &
                                      'MPI_GET_COUNT() returned ', &
                                      IERR
                                    CALL MPITEST_MESSAGE( &
                                       MPITEST_NONFATAL, &
                                       INFOBUF)
                                    INFOBUF=' '
                                    CALL MPI_ERROR_STRING(IERR, &
                                       INFOBUF, SIZE, IERR2)
                                    CALL MPITEST_MESSAGE( &
                                       MPITEST_FATAL, &
                                       INFOBUF)
                                 END IF

                                 IF (COUNT .NE. LENGTH) THEN
                                    FAIL = FAIL + 1
                                    WRITE(INFOBUF, 70) 'STATUS from ', &
                                      'MPI_IPROBE() has unexpected ', &
                                      'length Expected: ', LENGTH, &
                                      'Actual: ', COUNT
                                    CALL MPITEST_MESSAGE( &
                                       MPITEST_NONFATAL, &
                                       INFOBUF)
                                 END IF
#else
                                 WRITE(INFOBUF, 20) &
                                   'Output STATUS *not* verified'
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_NONINFO1, &
                                    INFOBUF)
#endif

                              END IF
 400                       CONTINUE
                        END IF

                        CALL MPI_BARRIER(COMM, IERR)
                        IF (IERR .NE. MPI_SUCCESS) THEN
                           WRITE(INFOBUF,10) &
                              'MPI_BARRIER() returned ', IERR 
                           CALL MPITEST_MESSAGE( &
                              MPITEST_NONFATAL, INFOBUF)
                           INFOBUF=' '
                           CALL MPI_ERROR_STRING(IERR, &
                              INFOBUF, SIZE, IERR2)
                           CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                              INFOBUF)
                        END IF

#ifdef MPITEST_BUFFER_RECV
                        IF (MPITEST_CURRENT_RANK .EQ. ROOT) THEN
                           DO 500 SOURCE=0, MAXNP - 1, 1
                              IF (SOURCE .NE. ROOT) THEN
                                 WRITE(INFOBUF, 60) &
                                    'Receiving from source ', &
                                    SOURCE, ' tag: ', TAG
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_INFO1, INFOBUF)

                                 CALL MPI_RECV(BUFFER, LENGTH, &
                                    MPITEST_MPI_DATATYPES &
                                       (TEST_TYPE), &
                                    SOURCE, TAG, COMM, STATUS, IERR)

                                 IF (IERR .NE. MPI_SUCCESS) THEN
                                    FAIL = FAIL + 1
                                    WRITE(INFOBUF, 10) &
                                       'MPI_RECV() returned ', &
                                       IERR
                                    CALL MPITEST_MESSAGE( &
                                       MPITEST_NONFATAL, &
                                       INFOBUF)
                                    INFOBUF=' '
                                    CALL MPI_ERROR_STRING(IERR, &
                                       INFOBUF, SIZE, IERR2)
                                    CALL MPITEST_MESSAGE( &
                                       MPITEST_FATAL, &
                                       INFOBUF)
                                 END IF

#ifdef MPITEST_BUFFER_CHK
                                 WRITE(INFOBUF, 20) &
                                    'Verifying data received'
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_INFO1, INFOBUF)

!                                Expecting to receive rank #
!                                from each rank
!
                                 CALL MPITEST_DATATEMPLATE_INIT( &
                                    VALUE, SOURCE, TEST_TYPE)

                                 CALL MPITEST_BUFFER_ERRORS( &
                                    TEST_TYPE, LENGTH, VALUE, &
                                    BUFFER, ERROR)

                                 IF (ERROR .NE. 0) THEN
                                    WRITE(INFOBUF, 80) &
                                       ERROR, ' errors in buffer'
                                    CALL MPITEST_MESSAGE( &
                                       MPITEST_NONFATAL, &
                                       INFOBUF)
                                    FAIL = FAIL + 1
                                 END IF
#else
                                 WRITE(INFOBUF, 20) &
                                    'data received *not* verified',
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_INFO1, INFOBUF)
#endif
                              END IF
 500                       CONTINUE
                        END IF
                        IF ((MPITEST_CURRENT_RANK .NE. ROOT) .AND. &
                            (MPITEST_CURRENT_RANK .LT. MAXNP)) &
                               THEN
!                          Wait on message sent to root from
!                          each rank
!
                           CALL MPI_WAIT(REQUEST, STATUS, IERR)
                           IF (IERR .NE. MPI_SUCCESS) THEN
                              FAIL = FAIL + 1
                              WRITE(INFOBUF, 10) &
                                'MPI_WAIT() returned ', &
                                IERR
                              CALL MPITEST_MESSAGE( &
                                 MPITEST_NONFATAL, &
                                 INFOBUF)
                              INFOBUF=' '
                              CALL MPI_ERROR_STRING(IERR, &
                                 INFOBUF, SIZE, IERR2)
                              CALL MPITEST_MESSAGE( &
                                 MPITEST_FATAL, INFOBUF)
                           END IF
                        END IF
#else
                        IF ((MPITEST_CURRENT_RANK .NE. ROOT) .AND. &
                            (MPITEST_CURRENT_RANK .LT. MAXNP)) &
                               THEN
!                          Cancel message sent to root from
!                          each rank
!
                           WRITE(INFOBUF, 60) &
                              'Cancelling messge from source ', &
                              MPITEST_current_rank, ' tag: ', &
                              TAG
                           CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                              INFOBUF)

                           CALL MPI_CANCEL(REQUEST, IERR)
                           IF (IERR .NE. MPI_SUCCESS) THEN
                              FAIL = FAIL + 1
                              WRITE(INFOBUF, 10) &
                                'MPI_CANCEL() returned ', &
                                IERR
                              CALL MPITEST_MESSAGE( &
                                 MPITEST_NONFATAL, &
                                 INFOBUF)
                              INFOBUF=' '
                              CALL MPI_ERROR_STRING(IERR, &
                                 INFOBUF, SIZE, IERR2)
                              CALL MPITEST_MESSAGE( &
                                 MPITEST_FATAL, INFOBUF)
                           END IF
                        END IF
#endif
                        CALL MPI_BARRIER(COMM, IERR)
                        IF (IERR .NE. MPI_SUCCESS) THEN
                           FAIL = FAIL + 1
                           WRITE(INFOBUF, 10) &
                              'MPI_BARRIER() returned ', &
                              IERR
                           CALL MPITEST_MESSAGE( &
                              MPITEST_NONFATAL, &
                              INFOBUF)
                           INFOBUF=' '
                           CALL MPI_ERROR_STRING(IERR, &
                              INFOBUF, SIZE, IERR2)
                           CALL MPITEST_MESSAGE( &
                              MPITEST_FATAL, INFOBUF)
                        END IF
                     END IF
 300              CONTINUE

#ifdef MPITEST_STATUS_CHK
 200           CONTINUE
#endif

            END IF
         END IF

 800     CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)
 100  CONTINUE

      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END

     
