! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
!
!                          MPI_Graph_create_err5
!
!   Reference:  MPI Standard,  Chapter 6:  Process Topologies
!                              Section 6.5.3:  General(Graph) Constructor
!
!   MPI_Graph_create error tests
!   -----------------------------------
!  1)  Nodes < 0 ......................................... [ARG/OTHER]
!  2)  Call with nodes > MPI_COMM_WORLD .................. [ARG/OTHER]
!  3)  Incompatible parameters:  Edge number too large ... [ARG/TOPOLOGY/OTHER]
!  4)  Incompatible parameters:  Negative edge number .... [ARG/TOPOLOGY/OTHER]
!  5)  Incompatible parameters:  Null edge ............... [ARG/TOPOLOGY/OTHER]
!  6)  Call with MPI_COMM_NULL ........................... [COMM]
!
!***********************************************************************

#include "f90options.h"

      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
      INTEGER    MPITEST_DIMS
      PARAMETER( MPITEST_DIMS = 2)
      CHARACTER  TEST_DESC*(*)
      PARAMETER( TEST_DESC = 'MPI_GRAPH_CREATE: Null Edge') 
!                           counts total number of PASSes/FAILures/VERIFY
      INTEGER PASS
      INTEGER FAIL              
!                           return value from MPI calls
      INTEGER IERR
      INTEGER ERR
!                           error class of IERR
      INTEGER ERRORCLASS
      INTEGER SIZE 

      INTEGER I, J
!                           general loop indices
      INTEGER NUM_DIMS
!                           number of dimensions to generate coordinates for
      INTEGER TEST_NUMP
!                           the number of processors in current communicator  
      INTEGER NNODES
!                           number of nodes in graph
      INTEGER RANK_COMM
!                           rank in current communicator
      LOGICAL REORDER
!                           flag telling if ranks can be reordered
      CHARACTER*(IOSIZE)  INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
      CHARACTER*32   TESTNAME
!                           name of this test
      INTEGER COMM_GRAPH
!                           graph communicator
      INTEGER INDEX(4)
!                           graph cumulative connection count array
      INTEGER  EDGES(6)
!                           graph node neighbors list array
      LOGICAL PERIODS(MPITEST_DIMS)
!                           flag to allow periodic assignment
!-----------------------------------------------------------------------------
 98   FORMAT(A,A)
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT)
 102  FORMAT(A, INT_FMT, A, INT_FMT, A, INT_FMT, A )
 103  FORMAT(A, INT_FMT, A, A)
!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
      TESTNAME='MPI_GRAPH_CREATE_err5'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      PASS = 0
      FAIL = 0
! 
!  Set an errorhandler so we get control back.  
! 
      CALL MPI_ERRHANDLER_SET(MPI_COMM_WORLD, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 99) ' MPI_ERRORHANDLER_SET returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
      CALL MPI_COMM_SIZE(MPI_COMM_WORLD, TEST_NUMP, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
        WRITE(INFOBUF,102) 'MPI_COMM_SIZE Error (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!  We need 4 nodes
!
      IF(TEST_NUMP .LT. 4) THEN
      WRITE(INFOBUF,103)  'Ranks =', TEST_NUMP, ' MPI_GRAPH_CREATE, ', &
                                                   'needs at least 4'
      CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!--------------------------  MPII_GRAPH_CREATE  ------------------------
!
      NUM_DIMS = 2
      NNODES   = 4

      EDGES(1) = 1
      EDGES(2) = 3
      EDGES(3) = 0
      EDGES(4) = 3
      EDGES(5) = 0
      EDGES(6) = 2

      INDEX(1) = 2
      INDEX(2) = 3
      INDEX(3) = 4
      INDEX(4) = 6
!  Make a null edge
      EDGES(1) = 0
!
      reorder = .TRUE.

      CALL  MPI_GRAPH_CREATE(MPI_COMM_WORLD, &
                             NNODES, &
                             INDEX, &
                             EDGES, &
                             REORDER, &
                             COMM_GRAPH, &
                             IERR)

      IF (IERR .EQ. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 98) TEST_DESC, ' did not FAIL'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      ELSE
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
        IF (ERRORCLASS .NE. MPI_ERR_ARG       .AND.    &
            ERRORCLASS .NE. MPI_ERR_TOPOLOGY  .AND.    &
            ERRORCLASS .NE. MPI_ERR_OTHER)  THEN

          FAIL = FAIL + 1
          WRITE(INFOBUF,101) TEST_DESC, ' Returned',  ERRORCLASS 
!
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          WRITE(INFOBUF,102)  &
            'Expected MPI_ERR_ARG(',       MPI_ERR_ARG,   &
              ')  OR  MPI_ERR_TOPOLOGY(',  MPI_ERR_TOPOLOGY,  &
              ')  OR  MPI_ERR_OTHER(',     MPI_ERR_OTHER, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ELSE
          PASS = PASS + 1
        END IF
      END IF
!
      CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
!
!  REPORT OVERALL RESULTS
!
      CALL MPITEST_REPORT(PASS, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END
